import os
import sys
import unittest
import numpy as np
from scipy.optimize import approx_fprime

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import residual_thy, residual_voigt
from TRXASprefitpack import res_grad_voigt, res_grad_thy
from TRXASprefitpack import voigt, voigt_thy
from TRXASprefitpack import edge_gaussian, edge_lorenzian

rel = 1e-4
epsilon = 1e-7

class TestGradResStatic(unittest.TestCase):

    def test_res_grad_voigt_1(self):
        '''
        Gaussian Edge
        '''
        e0_1 = 8987; e0_2 = 9000
        e0_edge = 8992
        fwhm_G_1 = 0.8
        fwhm_G_2 = 0.9
        fwhm_L_1 = 3
        fwhm_L_2 = 9
        fwhm_edge = 7
        
        # set scan range
        e = np.linspace(8960, 9020, 160)
        
        # generate model spectrum
        model_static = 0.1*voigt(e-e0_1, fwhm_G_1, fwhm_L_1) + \
            0.7*voigt(e-e0_2, fwhm_G_2, fwhm_L_2) + \
                0.2*edge_gaussian(e-e0_edge, fwhm_edge)
        eps_static = np.ones_like(model_static)

        x0 = [8985, 8997, 0.7, 0.8, 2, 7, 8990, 9]

        ref_res = np.sum(residual_voigt(x0, 2, 'g', 1, None,
        e=e, intensity=model_static, eps=eps_static)**2)/2

        ref_grad = approx_fprime(x0, lambda x0: 
        np.sum(residual_voigt(x0, 2, 'g', 1, None,
        e=e, intensity=model_static, eps=eps_static)**2)/2, 1e-7)

        res_tst, grad_tst = res_grad_voigt(x0, 2, 'g', 1, 
        None, np.zeros(len(x0), dtype=bool),
        e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_voigt_2(self):
        '''
        Lorenzian Edge
        '''
        e0_1 = 8987; e0_2 = 9000
        e0_edge = 8992
        fwhm_G_1 = 0.8
        fwhm_G_2 = 0.9
        fwhm_L_1 = 3
        fwhm_L_2 = 9
        fwhm_edge = 7
        
        # set scan range
        e = np.linspace(8960, 9020, 160)
        
        # generate model spectrum
        model_static = 0.1*voigt(e-e0_1, fwhm_G_1, fwhm_L_1) + \
            0.7*voigt(e-e0_2, fwhm_G_2, fwhm_L_2) + \
                0.2*edge_lorenzian(e-e0_edge, fwhm_edge)
        eps_static = np.ones_like(model_static)

        x0 = [8985, 8997, 0.7, 0.8, 2, 7, 8990, 9]

        ref_res = np.sum(residual_voigt(x0, 2, 'l', 1, None,
        e=e, intensity=model_static, eps=eps_static)**2)/2

        ref_grad = approx_fprime(x0, lambda x0: 
        np.sum(residual_voigt(x0, 2, 'l', 1, None,
        e=e, intensity=model_static, eps=eps_static)**2)/2, 1e-7)

        res_tst, grad_tst = res_grad_voigt(x0, 2, 'l', 1, 
        None, np.zeros(len(x0), dtype=bool),
        e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_voigt_3(self):
        '''
        Gaussian Edge + 2nd order polynomial baseline
        '''
        e0_1 = 8987; e0_2 = 9000
        e0_edge = 8992
        fwhm_G_1 = 0.8
        fwhm_G_2 = 0.9
        fwhm_L_1 = 3
        fwhm_L_2 = 9
        fwhm_edge = 7
        
        # set scan range
        e = np.linspace(8960, 9020, 160)
        
        # generate model spectrum
        model_static = 0.1*voigt(e-e0_1, fwhm_G_1, fwhm_L_1) + \
            0.7*voigt(e-e0_2, fwhm_G_2, fwhm_L_2) + \
                0.2*edge_gaussian(e-e0_edge, fwhm_edge)+\
                3e-5*(e-e0_edge)**2+1e-3*(e-e0_edge)+1e-1
        eps_static = np.ones_like(model_static)

        x0 = [8985, 8997, 0.7, 0.8, 2, 7, 8990, 9]

        ref_res = np.sum(residual_voigt(x0, 2, edge='g', num_edge=1, base_order=2,
        e=e, intensity=model_static, eps=eps_static)**2)/2

        ref_grad = approx_fprime(x0, lambda x0: 
        np.sum(residual_voigt(x0, 2, 'g', 1, 2,
        e=e, intensity=model_static, eps=eps_static)**2)/2, 1e-6)

        res_tst, grad_tst = res_grad_voigt(x0, 2, 'g', 1, 
        2, np.zeros(len(x0), dtype=bool),
        e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_voigt_4(self):
        '''
        Lorenzian Edge + 2nd order polynomial baseline
        '''
        e0_1 = 8987; e0_2 = 9000
        e0_edge = 8992
        fwhm_G_1 = 0.8
        fwhm_G_2 = 0.9
        fwhm_L_1 = 3
        fwhm_L_2 = 9
        fwhm_edge = 7
        
        # set scan range
        e = np.linspace(8960, 9020, 160)
        
        # generate model spectrum
        model_static = 0.1*voigt(e-e0_1, fwhm_G_1, fwhm_L_1) + \
            0.7*voigt(e-e0_2, fwhm_G_2, fwhm_L_2) + \
                0.2*edge_lorenzian(e-e0_edge, fwhm_edge)+\
                3e-5*(e-e0_edge)**2+1e-3*(e-e0_edge)+1e-1
        eps_static = np.ones_like(model_static)

        x0 = [8985, 8997, 0.7, 0.8, 2, 7, 8990, 9]

        ref_res = np.sum(residual_voigt(x0, 2, edge='l', num_edge=1, base_order=2,
        e=e, intensity=model_static, eps=eps_static)**2)/2

        ref_grad = approx_fprime(x0, lambda x0: 
        np.sum(residual_voigt(x0, 2, 'l', 1, 2,
        e=e, intensity=model_static, eps=eps_static)**2)/2, 1e-6)

        res_tst, grad_tst = res_grad_voigt(x0, 2, 'l', 1, 
        2, np.zeros(len(x0), dtype=bool),
        e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_thy_1(self):
        '''
        two thy component and two gaussian edge and 2nd order polynomial baseline (policy shift)
        '''

        e0_edge = np.array([860.5, 862])
        fwhm_edge = np.array([1, 1.5])
        peak_shift = np.array([862.5, 863])
        mixing = np.array([0.3, 0.7])
        mixing_edge = np.array([0.3, 0.7])
        fwhm_G_thy = 0.3
        fwhm_L_thy = 0.5
        
        thy_peak = np.empty(2, dtype=object)
        thy_peak[0] = np.genfromtxt(path+'/'+'Ni_example_1.stk')
        thy_peak[1] = np.genfromtxt(path+'/'+'Ni_example_2.stk')
        
        # set scan range
        e = np.linspace(852.5, 865, 51)
        
        # generate model spectrum
        model_static = mixing[0]*voigt_thy(e, thy_peak[0], fwhm_G_thy, fwhm_L_thy,
        peak_shift[0], policy='shift')+\
            mixing[1]*voigt_thy(e, thy_peak[1], fwhm_G_thy, fwhm_L_thy,
            peak_shift[1], policy='shift')+\
                mixing_edge[0]*edge_gaussian(e-e0_edge[0], fwhm_edge[0])+\
                    mixing_edge[1]*edge_gaussian(e-e0_edge[1], fwhm_edge[1])+\
                    1e-2*(e-860)**2 + 2e-1*(e-860) + 3e-1
        
        eps_static = np.ones_like(model_static)

        x0 = [0.2, 0.4, 862.2, 863.3, 860, 861.5, 0.8, 1.2]

        ref_res = np.sum(residual_thy(x0, 'shift', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2)/2
        ref_grad = approx_fprime(x0, lambda x0: 1/2*np.sum(residual_thy(x0, 'shift', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2), 1e-6)

        res_tst, grad_tst = res_grad_thy(x0, 'shift', thy_peak, 'g', 2, 2,
        np.zeros(len(x0), dtype=bool), e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_thy_2(self):
        '''
        two thy component and two gaussian edge and 2nd order polynomial baseline (policy scale)
        '''

        e0_edge = np.array([860.5, 862])
        fwhm_edge = np.array([1, 1.5])
        peak_shift = np.array([862.5, 863])
        mixing = np.array([0.3, 0.7])
        mixing_edge = np.array([0.3, 0.7])
        fwhm_G_thy = 0.3
        fwhm_L_thy = 0.5
        
        thy_peak = np.empty(2, dtype=object)
        thy_peak[0] = np.genfromtxt(path+'/'+'Ni_example_1.stk')
        thy_peak[1] = np.genfromtxt(path+'/'+'Ni_example_2.stk')
        
        # set scan range
        e = np.linspace(852.5, 865, 51)
        
        # generate model spectrum
        model_static = mixing[0]*voigt_thy(e, thy_peak[0], fwhm_G_thy, fwhm_L_thy,
        peak_shift[0], policy='shift')+\
            mixing[1]*voigt_thy(e, thy_peak[1], fwhm_G_thy, fwhm_L_thy,
            peak_shift[1], policy='shift')+\
                mixing_edge[0]*edge_gaussian(e-e0_edge[0], fwhm_edge[0])+\
                    mixing_edge[1]*edge_gaussian(e-e0_edge[1], fwhm_edge[1])+\
                    1e-2*(e-860)**2 + 2e-1*(e-860) + 3e-1
        
        eps_static = np.ones_like(model_static)

        x0 = [0.2, 0.4, 1.01, 1.05, 1.01, 860, 862.5, 0.8, 1.2]

        ref_res = np.sum(residual_thy(x0, 'scale', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2)/2
        ref_grad = approx_fprime(x0, lambda x0: 1/2*np.sum(residual_thy(x0, 'scale', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2), 1e-6)

        res_tst, grad_tst = res_grad_thy(x0, 'scale', thy_peak, 'g', 2, 2,
        np.zeros(len(x0), dtype=bool), e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_thy_3(self):
        '''
        two thy component and two gaussian edge and 2nd order polynomial baseline (policy both)
        '''

        e0_edge = np.array([860.5, 862])
        fwhm_edge = np.array([1, 1.5])
        peak_shift = np.array([862.5, 863])
        mixing = np.array([0.3, 0.7])
        mixing_edge = np.array([0.3, 0.7])
        fwhm_G_thy = 0.3
        fwhm_L_thy = 0.5
        
        thy_peak = np.empty(2, dtype=object)
        thy_peak[0] = np.genfromtxt(path+'/'+'Ni_example_1.stk')
        thy_peak[1] = np.genfromtxt(path+'/'+'Ni_example_2.stk')
        
        # set scan range
        e = np.linspace(852.5, 865, 51)
        
        # generate model spectrum
        model_static = mixing[0]*voigt_thy(e, thy_peak[0], fwhm_G_thy, fwhm_L_thy,
        peak_shift[0], policy='shift')+\
            mixing[1]*voigt_thy(e, thy_peak[1], fwhm_G_thy, fwhm_L_thy,
            peak_shift[1], policy='shift')+\
                mixing_edge[0]*edge_gaussian(e-e0_edge[0], fwhm_edge[0])+\
                    mixing_edge[1]*edge_gaussian(e-e0_edge[1], fwhm_edge[1])+\
                    1e-2*(e-860)**2 + 2e-1*(e-860) + 3e-1
        
        eps_static = np.ones_like(model_static)

        x0 = [0.2, 0.4, 862.2, 863.3, 1.01, 1.05, 860.5, 862.5, 0.8, 1.2]

        ref_res = np.sum(residual_thy(x0, 'both', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2)/2
        ref_grad = approx_fprime(x0, lambda x0: 1/2*np.sum(residual_thy(x0, 'both', thy_peak,
        'g', 2, 2, e=e, intensity=model_static, eps=eps_static)**2), 1e-6)

        res_tst, grad_tst = res_grad_thy(x0, 'both', thy_peak, 'g', 2, 2,
        np.zeros(len(x0), dtype=bool), e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

    def test_res_grad_thy_4(self):
        '''
        two thy component and two lorenzian edge and 2nd order polynomial baseline (policy both)
        '''

        e0_edge = np.array([860.5, 862])
        fwhm_edge = np.array([1, 1.5])
        peak_shift = np.array([862.5, 863])
        mixing = np.array([0.3, 0.7])
        mixing_edge = np.array([0.3, 0.7])
        fwhm_G_thy = 0.3
        fwhm_L_thy = 0.5
        
        thy_peak = np.empty(2, dtype=object)
        thy_peak[0] = np.genfromtxt(path+'/'+'Ni_example_1.stk')
        thy_peak[1] = np.genfromtxt(path+'/'+'Ni_example_2.stk')
        
        # set scan range
        e = np.linspace(852.5, 865, 51)
        
        # generate model spectrum
        model_static = mixing[0]*voigt_thy(e, thy_peak[0], fwhm_G_thy, fwhm_L_thy,
        peak_shift[0], policy='shift')+\
            mixing[1]*voigt_thy(e, thy_peak[1], fwhm_G_thy, fwhm_L_thy,
            peak_shift[1], policy='shift')+\
                mixing_edge[0]*edge_lorenzian(e-e0_edge[0], fwhm_edge[0])+\
                    mixing_edge[1]*edge_lorenzian(e-e0_edge[1], fwhm_edge[1])+\
                    1e-2*(e-860)**2 + 2e-1*(e-860) + 3e-1
        
        eps_static = np.ones_like(model_static)

        x0 = [0.2, 0.4, 862.2, 863.3, 1.01, 1.05, 860.5, 862.5, 0.8, 1.2]

        ref_res = np.sum(residual_thy(x0, 'both', thy_peak,
        'l', 2, 2, e=e, intensity=model_static, eps=eps_static)**2)/2
        ref_grad = approx_fprime(x0, lambda x0: 1/2*np.sum(residual_thy(x0, 'both', thy_peak,
        'l', 2, 2, e=e, intensity=model_static, eps=eps_static)**2), 1e-6)

        res_tst, grad_tst = res_grad_thy(x0, 'both', thy_peak, 'l', 2, 2,
        np.zeros(len(x0), dtype=bool), e=e, intensity=model_static, eps=eps_static)

        cond_res = np.allclose(res_tst, ref_res)
        cond_grad = np.allclose(grad_tst, ref_grad, rtol=1e-4, atol=1e-6)

        self.assertEqual((cond_res, cond_grad), (True, True))

if __name__ == '__main__':
    unittest.main()      


import os
import sys
import unittest
import numpy as np
from scipy.optimize import approx_fprime

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import residual_decay, residual_dmp_osc, residual_both
from TRXASprefitpack import res_grad_decay, res_grad_dmp_osc, res_grad_both
from TRXASprefitpack import solve_seq_model, rate_eq_conv
from TRXASprefitpack import dmp_osc_conv

rel = 1e-4
epsilon = 1e-7

class TestGradResTransient(unittest.TestCase):

    def test_res_grad_decay_1(self):
        '''
        Test res_grad_decay (irf: g)
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]
        abs_2 = [0.5, 0.8, 0.2, 0]
        abs_3 = [-0.5, 0.7, 0.9, 0]
        abs_4 = [0.6, 0.3, -1, 0]
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, irf='g')
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, irf='g')
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, irf='g')
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, irf='g')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2
        i_obs_3 = y_obs_3
        i_obs_4 = y_obs_4

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.4, 9, 990])

        res_ref = 1/2*np.sum(residual_decay(x0, False, 'g', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_decay(x0, False, 'g', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_decay(x0, 3, False, 'g', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_decay_2(self):
        '''
        Test res_grad_decay (irf: c)
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]
        abs_2 = [0.5, 0.8, 0.2, 0]
        abs_3 = [-0.5, 0.7, 0.9, 0]
        abs_4 = [0.6, 0.3, -1, 0]
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, irf='c')
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, irf='c')
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, irf='c')
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, irf='c')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2
        i_obs_3 = y_obs_3 
        i_obs_4 = y_obs_4 

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.4, 9, 990])

        res_ref = 1/2*np.sum(residual_decay(x0, False, 'c', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_decay(x0, False, 'c', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_decay(x0, 3, False, 'c', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_decay_3(self):
        '''
        Test res_grad_decay (irf: pv)
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100; eta = 0.7
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]
        abs_2 = [0.5, 0.8, 0.2, 0]
        abs_3 = [-0.5, 0.7, 0.9, 0]
        abs_4 = [0.6, 0.3, -1, 0]
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2 
        i_obs_3 = y_obs_3 
        i_obs_4 = y_obs_4 

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.3, 0.15, 0, 0, 0, 0, 0.4, 9, 990])

        res_ref = 1/2*np.sum(residual_decay(x0, True, 'pv', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_decay(x0, True, 'pv', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_decay(x0, 3, True, 'pv', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))
    
    def test_res_grad_dmp_osc_1(self):
        '''
        Gaussian IRF
        '''
        fwhm = 0.100
        tau = np.array([0.5, 10, 1000])
        period = np.array([0.2, 3, 200])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = np.array([1, 1, 1])
        abs_2 = np.array([0.5, 0.8, 0.2])
        abs_3 = np.array([-0.5, 0.7, 0.9])
        abs_4 = np.array([0.6, 0.3, -1])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = dmp_osc_conv(t_seq-t0[0], fwhm, tau, period, phase, abs_1, irf='g')
        y_obs_2 = dmp_osc_conv(t_seq-t0[1], fwhm, tau, period, phase, abs_2, irf='g')
        y_obs_3 = dmp_osc_conv(t_seq-t0[2], fwhm, tau, period, phase, abs_3, irf='g')
        y_obs_4 = dmp_osc_conv(t_seq-t0[3], fwhm, tau, period, phase, abs_4, irf='g')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2
        i_obs_3 = y_obs_3
        i_obs_4 = y_obs_4

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.4, 9, 990, 0.3, 5, 250])

        res_ref = 1/2*np.sum(residual_dmp_osc(x0, 3, 'g', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_dmp_osc(x0, 3, 'g', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_dmp_osc(x0, 3, 'g', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_dmp_osc_2(self):
        '''
        Cauchy IRF
        '''
        fwhm = 0.100
        tau = np.array([0.5, 10, 1000])
        period = np.array([0.2, 3, 200])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = np.array([1, 1, 1])
        abs_2 = np.array([0.5, 0.8, 0.2])
        abs_3 = np.array([-0.5, 0.7, 0.9])
        abs_4 = np.array([0.6, 0.3, -1])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = dmp_osc_conv(t_seq-t0[0], fwhm, tau, period, phase, abs_1, irf='c')
        y_obs_2 = dmp_osc_conv(t_seq-t0[1], fwhm, tau, period, phase, abs_2, irf='c')
        y_obs_3 = dmp_osc_conv(t_seq-t0[2], fwhm, tau, period, phase, abs_3, irf='c')
        y_obs_4 = dmp_osc_conv(t_seq-t0[3], fwhm, tau, period, phase, abs_4, irf='c')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2
        i_obs_3 = y_obs_3
        i_obs_4 = y_obs_4

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.4, 9, 990, 0.3, 5, 250])

        res_ref = 1/2*np.sum(residual_dmp_osc(x0, 3, 'c', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_dmp_osc(x0, 3, 'c', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_dmp_osc(x0, 3, 'c', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_dmp_osc_3(self):
        '''
        Pseudo Voigt IRF
        '''
        fwhm = 0.100; eta = 0.7
        tau = np.array([0.5, 10, 1000])
        period = np.array([0.2, 3, 200])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = np.array([1, 1, 1])
        abs_2 = np.array([0.5, 0.8, 0.2])
        abs_3 = np.array([-0.5, 0.7, 0.9])
        abs_4 = np.array([0.6, 0.3, -1])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = dmp_osc_conv(t_seq-t0[0], fwhm, tau, period, phase, abs_1, 
        irf='pv', eta=eta)
        y_obs_2 = dmp_osc_conv(t_seq-t0[1], fwhm, tau, period, phase, abs_2, 
        irf='pv', eta=eta)
        y_obs_3 = dmp_osc_conv(t_seq-t0[2], fwhm, tau, period, phase, abs_3, 
        irf='pv', eta=eta)
        y_obs_4 = dmp_osc_conv(t_seq-t0[3], fwhm, tau, period, phase, abs_4, 
        irf='pv', eta=eta)
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1
        i_obs_2 = y_obs_2
        i_obs_3 = y_obs_3
        i_obs_4 = y_obs_4

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0.23, 0, 0, 0, 0, 0.4, 9, 990,  0.3, 5, 250])

        res_ref = 1/2*np.sum(residual_dmp_osc(x0, 3, 'pv', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_dmp_osc(x0, 3, 'pv', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_dmp_osc(x0, 3, 'pv', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon) # noise field

        self.assertEqual((check_res, check_grad), (True, True))


    def test_res_grad_both_1(self):
        '''
        Gaussian IRF
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100
        tau_osc = np.array([0.3, 8, 800])
        period_osc = np.array([0.2, 2, 300])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]; abs_1_osc = 0.5*np.array([0.6, 0.1, 0.3])
        abs_2 = [0.5, 0.8, 0.2, 0]; abs_2_osc = 0.1*np.array([0.2, 0.17, 1.19])
        abs_3 = [-0.5, 0.7, 0.9, 0]; abs_3_osc = -0.2*np.array([0.16, 0.28, 0.339])
        abs_4 = [0.6, 0.3, -1, 0]; abs_4_osc = 0.18*np.array([1, 3, -2])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, irf='g')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_1_osc, irf='g')
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, irf='g')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_2_osc, irf='g')
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, irf='g')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_3_osc, irf='g')
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, irf='g')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_4_osc, irf='g')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1 
        i_obs_2 = y_obs_2 
        i_obs_3 = y_obs_3 
        i_obs_4 = y_obs_4 

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.5, 9, 990, 1, 15, 1500, 0.5, 1, 200])

        res_ref = 1/2*np.sum(residual_both(x0, 3, 3, False, 'g', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_both(x0, 3, 3, False, 'g', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_both(x0, 3, 3, False, 'g', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon)

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_both_2(self):
        '''
        Cauchy IRF
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100
        tau_osc = np.array([0.3, 8, 800])
        period_osc = np.array([0.2, 2, 300])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]; abs_1_osc = 0.5*np.array([0.6, 0.1, 0.3])
        abs_2 = [0.5, 0.8, 0.2, 0]; abs_2_osc = 0.1*np.array([0.2, 0.17, 1.19])
        abs_3 = [-0.5, 0.7, 0.9, 0]; abs_3_osc = -0.2*np.array([0.16, 0.28, 0.339])
        abs_4 = [0.6, 0.3, -1, 0]; abs_4_osc = 0.18*np.array([1, 3, -2])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, irf='c')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_1_osc, irf='c')
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, irf='c')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_2_osc, irf='c')
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, irf='c')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_3_osc, irf='c')
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, irf='c')+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_4_osc, irf='c')
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1 
        i_obs_2 = y_obs_2 
        i_obs_3 = y_obs_3 
        i_obs_4 = y_obs_4 

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0, 0, 0, 0, 0.5, 9, 990, 1, 15, 1500, 0.5, 1, 200])

        res_ref = 1/2*np.sum(residual_both(x0, 3, 3, False, 'c', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_both(x0, 3, 3, False, 'c', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_both(x0, 3, 3, False, 'c', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon)

        self.assertEqual((check_res, check_grad), (True, True))

    def test_res_grad_both_3(self):
        '''
        Pseudo Voigt IRF
        '''
        tau_1 = 0.5; tau_2 = 10
        tau_3 = 1000; fwhm = 0.100; eta = 0.7
        tau_osc = np.array([0.3, 8, 800])
        period_osc = np.array([0.2, 2, 300])
        phase = np.array([np.pi/17, -np.pi/6, np.pi/2])
        # initial condition
        y0 = np.array([1, 0, 0, 0])
        
        # set time range (mixed step)
        t_seq1 = np.arange(-2, -1, 0.2)
        t_seq2 = np.arange(-1, 2, 0.02)
        t_seq3 = np.arange(2, 5, 0.2)
        t_seq4 = np.arange(5, 10, 1)
        t_seq5 = np.arange(10, 100, 10)
        t_seq6 = np.arange(100, 1000, 100)
        t_seq7 = np.linspace(1000, 2000, 2)
        
        t_seq = np.hstack((t_seq1, t_seq2, t_seq3, t_seq4, t_seq5, t_seq6, t_seq7))
        eigval_seq, V_seq, c_seq = solve_seq_model(np.array([tau_1, tau_2, tau_3]), y0)
        
        # Now generates measured transient signal
        # Last element is ground state
        abs_1 = [1, 1, 1, 0]; abs_1_osc = 0.5*np.array([0.6, 0.1, 0.3])
        abs_2 = [0.5, 0.8, 0.2, 0]; abs_2_osc = 0.1*np.array([0.2, 0.17, 1.19])
        abs_3 = [-0.5, 0.7, 0.9, 0]; abs_3_osc = -0.2*np.array([0.16, 0.28, 0.339])
        abs_4 = [0.6, 0.3, -1, 0]; abs_4_osc = 0.18*np.array([1, 3, -2])
        
        t0 = np.random.uniform(-0.2, 0.2, 4) # perturb time zero of each scan
        
        # generate measured data
        y_obs_1 = rate_eq_conv(t_seq-t0[0], fwhm, abs_1, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_1_osc, 
            irf='pv', eta=eta)
        y_obs_2 = rate_eq_conv(t_seq-t0[1], fwhm, abs_2, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_2_osc, 
            irf='pv', eta=eta)
        y_obs_3 = rate_eq_conv(t_seq-t0[2], fwhm, abs_3, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_3_osc, 
            irf='pv', eta=eta)
        y_obs_4 = rate_eq_conv(t_seq-t0[3], fwhm, abs_4, eigval_seq, V_seq, c_seq, 
        irf='pv', eta=eta)+\
            dmp_osc_conv(t_seq-t0[1], fwhm, tau_osc, period_osc, phase, abs_4_osc, 
            irf='pv', eta=eta)
        
        eps_obs_1 = np.ones_like(y_obs_1)
        eps_obs_2 = np.ones_like(y_obs_2)
        eps_obs_3 = np.ones_like(y_obs_3)
        eps_obs_4 = np.ones_like(y_obs_4)
        
        # generate measured intensity
        i_obs_1 = y_obs_1 
        i_obs_2 = y_obs_2 
        i_obs_3 = y_obs_3 
        i_obs_4 = y_obs_4 

        t = [t_seq] 
        intensity = [np.vstack((i_obs_1, i_obs_2, i_obs_3, i_obs_4)).T]
        eps = [np.vstack((eps_obs_1, eps_obs_2, eps_obs_3, eps_obs_4)).T]

        x0 = np.array([0.15, 0.3 , 0, 0, 0, 0, 0.5, 9, 990, 1, 15, 1500, 0.5, 1, 200])

        res_ref = 1/2*np.sum(residual_both(x0, 3, 3, False, 'pv', 
        t=t, intensity=intensity, eps=eps)**2)
        grad_ref = approx_fprime(x0, lambda x0: \
            1/2*np.sum(residual_both(x0, 3, 3, False, 'pv', 
            t=t, intensity=intensity, eps=eps)**2), epsilon)

        res_tst, grad_tst = res_grad_both(x0, 3, 3, False, 'pv', 
        np.zeros_like(x0, dtype=bool), t, intensity, eps)

        check_res = np.allclose(res_ref, res_tst)
        check_grad = np.allclose(grad_ref, grad_tst, rtol=rel,
        atol=epsilon)

        self.assertEqual((check_res, check_grad), (True, True))


if __name__ == '__main__':
    unittest.main()


'''
_ampgo:
Adaptive Memory Programing For Global Optimization
Based On Andrea Gavana's Implementation (see: http://infinity77.net/global_optimization/)

For implementation detail see the `Tabu Tunneling Method` section in the below paper.
http://leeds-faculty.colorado.edu/glover/fred%20pubs/416%20-%20AMP%20(TS)%20for%20Constrained%20Global%20Opt%20w%20Lasdon%20et%20al%20.pdf

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Callable, Optional
import numpy as np
from scipy.optimize import minimize, OptimizeResult

def ampgo(func: Callable, x0: np.ndarray,
          tot_iter: Optional[int] = 20, max_tunnel: Optional[int] = 5,
          tol_tunnel: Optional[float] = 1e-5,
          minimizer_kwargs: Optional[dict] = None,
          eps1: Optional[float] = 0.02, eps2: Optional[float] = 0.1,
          n_tabu: Optional[int] = 5, strategy: Optional[str] = 'farthest',
          disp: Optional[bool] = False) -> OptimizeResult:

          '''
          ampgo: Adaptive Memory Programming for Global Optimization 
           Based on Tabu Tunneling Method
          
          Args:
           func: Objective function. 
             Signiture of objective function: `f(x, *args)`
           x0: initial guess
           tot_iter: maximum number of global iteration
           max_tunnel: maximum number of tunneling phase
           tol_tunnel: Tolerance to determine whether tunneling phase is 
             successful or not. 
             If :math:`|f(x_{tunnel})}<(1+{tol})(|f(x_{best})|+{tol})` then such tunneling phase
             is regarded as successful phase.
           minimizer_kwargs: Extra keyword arguments to be passed to the local minimizer
            `scipy.optimize.minimize`. Some important options could be:
             
            * method (str): The minimization Method (default: `L-BFGS-B`)

            * args (tuple): The extra arguments passed to the objective function (`func`) and
               its derivatives (`jac`, `hess`)
             
            * jac: jacobian of objective function (see scipy.optimize.minimize)
            * hess: hessian of objective function (see scipy.optimize.minimize)
            * bounds (Sequence of Tuple): Boundary of variable (see scipy.optimize.minimize)
           eps1: Constant used to define aspiration value
           eps2: Perturbation factor used to move away from the latest local minimum
           n_tabu: size of tabulist
           strategy ({'farthest', 'oldest'}): The strategy to delete element of tabulist when
            the size of tabulist exceeds `n_tabu`.

            * `farthest`: Delete farthest point from the latest local minimum point
            * `oldest`: Delete oldest point
           disp: display level, If zero or `None`, then no output is printed on screen.
             If postive number then status messages are printed.
          
          
          Returns:
           The optimization results represented as a `scipy.OptimizeResult` object.
           The important attributes are
           
           * `x`: The solution of the optimization
           * `fun, jac, hess`: values of objective fuction, its gradient and hessian.
           * `success`: Whether or not the optimizer exited successfuly
           * `message`: Description of the cause of the termination
          '''
          return


def check_vaild(x_try, tabulist):
    '''
    Check random pertubation of latest local mimum point
    is vaild.
    '''
    if not tabulist:
        return True
    else:
        dist = np.sum((tabulist-x_try)**2, axis=1)
        min_dist = np.min(dist)
        return min_dist > 1e-16
    

def delete_element(x_local, tabulist, strategy):
    '''
    Delete element from tabulist
    '''
    if strategy == 'oldest':
        tabulist.pop(0)
    else:
        dist = np.sum((tabulist-x_local)**2, axis=1)
        idx = np.argmax(dist)
        tabulist.pop(idx)

    return tabulist


def ttf(x0, *args):
    '''
    Tabu Tunneling function
    '''
    func, aspiration, tabulist = args[:3]
    fun_args = ()
    if len(args)>4:
        fun_args = tuple(args[4:])
    
    numerator = (func(x0, *fun_args)-aspiration)**2
    denominator = 1
    for tabu in tabulist:
        denominator = denominator*np.linalg.norm(x0-tabu)
    return numerator/denominator

def grad_ttf(x0, *args):
    '''
    Gradient of Tabu Tunneling function
    '''
    func, jac, aspiration, tabulist = args[0:4]

    fun_args = ()
    if len(args) > 4:
        fun_args = tuple(args[4:])
    
    fval = func(x0, *fun_args) - aspiration
    numerator = fval**2
    grad_numerator = 2*fval*jac(x0, *fun_args)
    denominator = 1.0
    grad_denom = np.zeros_like(x0)

    for tabu in tabulist:
        diff = tabu-x0
        dist = np.linalg.norm(diff)
        denominator = denominator*dist
        grad_denom = grad_denom + diff/dist**2
    
    return (grad_numerator+numerator*grad_denom)/denominator


def fun_grad_ttf(x0, *args):
    '''
    pair of function value and its gradient of
    Tabu Tunneling function
    '''
    '''
    Tabu Tunneling function
    '''
    func, aspiration, tabulist = args[:3]
    fun_args = ()
    if len(args)>4:
        fun_args = tuple(args[4:])
    
    f_val, grad_val = func(x0, *fun_args)
    f_val = f_val-aspiration
    numerator = f_val**2
    grad_numerator = 2*f_val*grad_val
    denominator = 1
    grad_denominator = np.zeros_like(x0)
    for tabu in tabulist:
        diff = tabu-x0
        dist = np.linalg.norm(diff)
        denominator = denominator*dist
        grad_denominator = grad_denominator + diff/dist**2
    
    y_ttf = numerator/denominator
    deriv_y_ttf = grad_numerator/denominator+y_ttf*grad_denominator
    return y_ttf, deriv_y_ttf


import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import exp_conv, fact_anal_exp_conv
from TRXASprefitpack import dmp_osc_conv, fact_anal_dmp_osc_conv
from TRXASprefitpack import sum_exp_dmp_osc_conv, fact_anal_sum_exp_dmp_osc_conv

class TestFactAnal(unittest.TestCase):

    def test_fact_anal_exp_conv(self):
        fwhm = 0.15
        c_ref = np.array([1, -0.5, -0.25])
        tau = np.array([0.3, 10])
        t = np.hstack((np.arange(-1, 1, 0.05), np.linspace(1, 100, 99)))
        model = exp_conv(t, fwhm, tau, c_ref, True, irf='g')
        eps = np.ones_like(model)
        expt = model
        c_tst = fact_anal_exp_conv(t, fwhm, tau, True, irf='g', intensity=expt, eps=eps)
        cond = np.allclose(c_ref, c_tst)
        self.assertEqual(cond, True)

    def test_fact_anal_dmp_osc_conv(self):
        fwhm = 0.15
        c_ref = np.array([1, 0.5])
        tau = np.array([0.3, 1])
        period = np.array([0.5, 1])
        phase = np.array([-np.pi/17, 2*np.pi/4])
        t = np.hstack((np.arange(-1, 1, 0.02), np.linspace(1, 100, 99)))
        model = dmp_osc_conv(t, fwhm, tau, period, phase, c_ref, irf='g')
        eps = np.ones_like(model)
        expt = model
        phase_tst, c_tst = fact_anal_dmp_osc_conv(t, fwhm, tau, period, irf='g', intensity=expt, eps=eps)
        cond = np.allclose(c_ref, c_tst)
        cond_phase = np.allclose(phase, phase_tst)
        self.assertEqual((cond, cond_phase), (True, True))

    def test_fact_anal_sum_exp_dmp_osc_conv(self):
        fwhm = 0.15
        c_ref_decay = np.array([1, -0.5, -0.25])
        c_ref_osc = np.array([0.15, 0.075])
        tau = np.array([0.5, 10])
        tau_osc = np.array([0.3, 1])
        period_osc = np.array([0.5, 1])
        phase_osc = np.array([np.pi/3, -np.pi/4])
        t = np.hstack((np.arange(-1, 1, 0.02), np.linspace(1, 100, 99)))
        model = sum_exp_dmp_osc_conv(t, fwhm, tau, tau_osc, period_osc, phase_osc, 
        c_ref_decay, c_ref_osc, base=True, irf='g')
        eps = np.ones_like(model)
        expt = model
        c_tst_decay, phase_tst_osc, c_tst_osc = \
            fact_anal_sum_exp_dmp_osc_conv(t, fwhm, tau, tau_osc, period_osc, base=True,
            irf='g', intensity=expt, eps=eps)
        cond_decay = np.allclose(c_ref_decay, c_tst_decay)
        cond_osc = np.allclose(c_ref_osc, c_tst_osc)
        cond_phase_osc = np.allclose(phase_osc, phase_tst_osc)
        self.assertEqual((cond_decay, cond_phase_osc, cond_osc), (True, True, True))


if __name__ == '__main__':
    unittest.main()


